/*
 *****************************************************
 *
 *  SaVi by Lloyd Wood (lloydwood@users.sourceforge.net),
 *          Patrick Worfolk (worfolk@alum.mit.edu) and
 *          Robert Thurman.
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  Also Copyright (c) 2006 by Lloyd Wood.
 *
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained from:
 *  http://savi.sourceforge.net/
 *  http://www.geom.uiuc.edu/locate/SaVi
 *
 *****************************************************
 *
 * fisheye.c
 *
 * $Id: fisheye.c,v 1.23 2006/07/18 13:22:22 lloydwood Exp $
 */

#include <math.h>
#include <stdlib.h>

#include "constants.h"
#include "globals.h"
#include "sats.h"
#include "tcl_utils.h"

#include "fisheye.h"

static unsigned int fisheye_flag = FALSE;
static unsigned int fisheye_labels_flag = FALSE;
static unsigned int fisheye_reverse_view_flag = FALSE;
static unsigned int fisheye_print_inview_flag = FALSE;
static double lat, lon;

static char plotStr[LENGTH_STRING_BUFFER];
static char fisheye_widget[] = ".fisheye.cmd.p";
static char font[] = "helvetica 9";

/*
 * fisheye_on.c
 *
 * Set flags to indicate that fisheye should be displayed.
 * Call fisheye_display to get updated display.
 */
char *
fisheye_on_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite_list SL = pconstellation->satellites;
  CentralBody *pcb = pconstellation->pcb;

  /* set variables so that fisheye will be displayed */
  fisheye_flag = TRUE;

  /* display the fisheye view */
  fisheye_display(SL, TRUE, pcb);

  return EMPTY_str;
}

char *
fisheye_off_cmd(int argc, char *argv[])
{
  fisheye_flag = FALSE;
  return EMPTY_str;
}

char *
fisheye_labels_on_cmd(int argc, char *argv[])
{
  fisheye_labels_flag = TRUE;
  return EMPTY_str;
}

char *
fisheye_labels_off_cmd(int argc, char *argv[])
{
  fisheye_labels_flag = FALSE;
  return EMPTY_str;
}

char *
fisheye_reverse_view_on_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite_list SL = pconstellation->satellites;
  CentralBody *pcb = pconstellation->pcb;

  fisheye_reverse_view_flag = TRUE;

  /* display the fisheye view */
  fisheye_display(SL, TRUE, pcb);

  return EMPTY_str;
}

char *
fisheye_reverse_view_off_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite_list SL = pconstellation->satellites;
  CentralBody *pcb = pconstellation->pcb;

  fisheye_reverse_view_flag = FALSE;

  /* display the fisheye view */
  if (fisheye_flag) fisheye_display(SL, TRUE, pcb);

  return EMPTY_str;
}

char *
fisheye_print_inview_on_cmd(int argc, char *argv[])
{
  fisheye_print_inview_flag = TRUE;

  return EMPTY_str;
}

char *
fisheye_print_inview_off_cmd(int argc, char *argv[])
{
  fisheye_print_inview_flag = FALSE;

  /* could delete or hide mask elevation circle */

  return EMPTY_str;
}

/*
 * fisheye_set
 *
 * Sets the lat and lon for the fisheye display and redisplays it
 */
char *
fisheye_set_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite_list sl = pconstellation->satellites;
  CentralBody *pcb = pconstellation->pcb;
  double new_lat = (double) atoi(argv[2]);
  double new_lon = (double) atoi(argv[4]);	/* was 3 gjl */
  const char format1[] = "set fisheye_lat_lon \"%d %.4f %c  %d %.4f %c\"";

  int lat_deg, lon_deg;
  char latNS, lonEW;
  double lat_min, lon_min;

  new_lat = (new_lat >= 0) ? new_lat + atof(argv[3]) / 60.0 :
    new_lat - atof(argv[3]) / 60.0;
  new_lon = (new_lon >= 0) ? new_lon + atof(argv[5]) / 60.0 :
    new_lon - atof(argv[5]) / 60.0;

  lat = new_lat;
  lon = new_lon;

  /* display lat lon */
  if (lat >= 0) {
    latNS = 'N';
    lat_deg = (int) floor(lat);
    lat_min = (lat - lat_deg) * 60.0;
  } else {
    latNS = 'S';
    lat_deg = (int) floor(-lat);
    lat_min = (-lat - lat_deg) * 60.0;
  }
  if (lon >= 0) {
    lonEW = 'E';
    lon_deg = (int) floor(lon);
    lon_min = (lon - lon_deg) * 60.0;
  } else {
    lonEW = 'W';
    lon_deg = (int) floor(-lon);
    lon_min = (-lon - lon_deg) * 60.0;
  }

  sprintf(cmd, format1, lat_deg, lat_min, latNS, lon_deg, lon_min, lonEW);
  tcl_script(cmd);

  /* display sky view */
  if (fisheye_flag) fisheye_display(sl, TRUE, pcb);
  return EMPTY_str;
}

/*
 * fisheye_display
 *
 * update the Tk image
 */
void
fisheye_display(const Satellite_list SL, unsigned int force,
		const CentralBody * pcb)
{
  int r;
  int offset, test, not_visible_to_terminal;
  float coverage_pct = (90.0 - coverage_angle) / 90.0;

  if (!fisheye_flag && !force)
    return;

  sprintf(cmd, "fisheye(delete)");
  test = tcl_script(cmd);
  if (debug && !test) {
    error("Problem telling Tk to delete fisheye.");
  }

  /* plot satellite positions */
  {
    CartesianCoordinates s;
    double sky[2];
    double t = current_time();
    double interval_time;
    Satellite_list sl = SL;

    unsigned int sats_visible, sats_above_mask;
    int d, x, y, x_start, y_start, x_end, y_end;
    int adjust = 2;

    sprintf(cmd,"fisheye(resize)");
    test = tcl_script(cmd);
    if (debug && !test) {
      error("problem telling Tk to resize fisheye disk.");
    }

    d = Fisheye_Diameter - adjust;
    r = d / 2 - 1;
    offset = r + adjust; /* important for later */
    
    if (coverage_type == MASK_ELEVATION) {
      x_start = d * coverage_angle / 180 + adjust;
      y_end = d - x_start;
      sprintf(cmd,"%s create oval %i %i %i %i -outline red -width 1 -tags mask_elevation", fisheye_widget, x_start, x_start, y_end, y_end);
      test = tcl_script(cmd);
      if (debug && !test) {
        error("Problem telling Tk to draw mask elevation.");
      }
    }

    sats_visible = sats_above_mask = 0;

    while (sl) {

      if (!sl->s->can_display_satellite) {
	sl = sl->next;
	continue;
      }

      oe_time_to_geocentric(&s, t, &(sl->s->oe), pcb);
      if (sat_to_fisheye(&s, lat, lon, pcb, t, sky)) {
	sats_visible++;
	if ((sky[0] * sky[0] + sky[1] * sky[1]) <
	    (coverage_pct * coverage_pct)) {
	  if (sl->s->gs_into_view == 0) {
	    sl->s->gs_into_view = t;
	    sl->s->gs_last_into_view = t;
	  }
	} else {
	  /*
	   * could add delta_t for half of interval before and after
	   * passing view test - but assumes delta_t unchanged!
	   */
	  if (sl->s->gs_into_view != 0) {
	    sl->s->gs_last_pass = t - sl->s->gs_into_view;
	  }
	  sl->s->gs_into_view = 0;
	}

	if (fisheye_reverse_view_flag) {
	  sky[0] = -sky[0];
	}
	not_visible_to_terminal = ((coverage_type == MASK_ELEVATION) &&
	                          (!sl->s->gs_into_view)) ||
	                          !sl->s->can_display_coverage;
	if (not_visible_to_terminal) {
	  sprintf(plotStr, "%d -fill red", sl->s->id);
	} else {
	  sats_above_mask++;
	  interval_time = t - sl->s->gs_into_view;
	  sprintf(plotStr, "{%d %.1f} -fill green", sl->s->id, interval_time);
	}
	x = (int) (offset + r * sky[0]);
	y = (int) (offset - r * sky[1]); /* North is top of canvas */

	x_start = x+6;
	y_start = y+6;

	if (fisheye_labels_flag) {
	  sprintf(cmd, "%s create text %i %i -text %s -font {%s} -tags labels",
		  fisheye_widget, x_start, y_start, plotStr, font);
	  test = tcl_script(cmd);
	  if (debug && !test) {
	    error("Problem telling Tk to write fisheye widget satellite labels.");
	  }
	}

	if (sl->s->tag == 1) {
	  sprintf(plotStr, "pink");
	} else if (not_visible_to_terminal) {
	  sprintf(plotStr, "red");
	} else {
	  sprintf(plotStr, "green");
	}

	x_start = x-3;
	x_end = x+4;
	y_start = y-3;
	y_end = y+4;

	sprintf(cmd, "%s create line %i %i %i %i -fill %s -tags crosses",
		fisheye_widget, x, y_start, x, y_end, plotStr);
	test = tcl_script(cmd);
	sprintf(cmd, "%s create line %i %i %i %i -fill %s -tags crosses",
                fisheye_widget, x_start, y, x_end, y, plotStr);
	test = test && tcl_script(cmd);
	if (debug && !test) {
	  error("Problem telling Tk to draw fisheye canvas crosses.");
	}
      } else if (sl->s->gs_last_pass != 0) {
	/*
	 * any recorded output needs formatting into a proper tracefile,
	 * including lat/long, estimated margin of error (delta_t/2).
	 * currently disabled.
	 */
	if (0) {
	  fprintf(stderr,
		  "Sat: %2d, first seen: %6.1f secs, for: %.1f secs.\n",
		  sl->s->id, sl->s->gs_last_into_view, sl->s->gs_last_pass);
	}
	sl->s->gs_into_view = 0;
      }

      sl = sl->next;
    }

    if (sats_above_mask && coverage_angle) {
      /* output number of satellites above a non-zero elevation angle */
      sprintf(plotStr, "\"%i in mask\"", sats_above_mask);
      sprintf(cmd, "%s create text %i %i -text %s -font {%s} -tags labels",
	      fisheye_widget, r, 10, plotStr, font);
      test = test && tcl_script(cmd);
    }

    /* output number of satellites visible */
    sprintf(plotStr, "\"%i in view\"", sats_visible);
    sprintf(cmd, "%s create text %i %i -text %s -font {%s} -tags labels",
	    fisheye_widget, r, d-9, plotStr, font);
    test = test && tcl_script(cmd);
    if (debug && !test) {
      error("Problem telling Tk to write satellite counts.");
    }
  }
}

